/* 
 * ihex2boot.c
 * Copyright (c) 2007 Hannu Nuotio <hannu.nuotio@pp.inet.fi>
 * Written by Hannu Nuotio <hannu.nuotio@pp.inet.fi>
 */

/*
 * ihex2boot converts a firmware file from Intel HEX to
 * the custom "boot" file format used by 1541-III-upgrade.
 *
 * Reset at location 0x000000 is copied to 0x00ff38 and patched to jump to the bootloader.
 * Bootloader at 0x00ff40-0x00ffff is protected by omitting any code at the address range.
 *
 * usage:
 *  ihex2boot < firmware.hex > fw.boot
 *
 */

#include <stdio.h>
#include <stdlib.h>

#define BOOT_HAVE_RESET	0x01
#define BOOT_DONE_RESET	0x02

unsigned char o[68]; // output buffer
int ocnt=0;
unsigned char icrc, outstate=0, resetstate=0;
unsigned char reset[4]; // output buffer

void outputhexblock() 
{
unsigned char ocrc,i;
// patch reset vector at 0x000000
if ((o[0]==0)&&(o[1]==0)&&(o[2]==0)) { // check if address 0x000000
	// save reset vector to reset[]
	for(i=0;i<4;i++) reset[i]=o[i+4];
	resetstate|=BOOT_HAVE_RESET;
	// reset vector = "goto 0xff40" (bootloader)
	o[4]=0xa0; o[5]=0xef; o[6]=0x7f; o[7]=0xf0;
	fprintf(stderr,"Reset vector at 0x000000 patched (block %i)\n",ocnt);
}
// patch start vector at 0x00ff38
if ((o[0]==0)&&(o[1]==0xff)&&(o[2]==0)) { // check if address 0x00ff00
	if (resetstate & BOOT_HAVE_RESET) { // check if reset vector is saved
		// start vector = old reset vector
		for(i=0;i<4;i++) o[4+0x38+i]=reset[i];
		resetstate|=BOOT_DONE_RESET;
		fprintf(stderr,"Start vector at 0x00ff38 patched (block %i)\n",ocnt);
	} else {
		// start vector = "goto 0x20" (guessed value)
		o[4+0x38]=0x10; o[5+0x38]=0xef; o[6+0x38]=0x00; o[7+0x38]=0xf0; 
		resetstate|=BOOT_DONE_RESET;
		fprintf(stderr,"WARNING! Patching \"goto 0x20\" to start vector at 0x00ff38 (block %i)\n",ocnt);
	}
}
// protect bootloader at 0x00ff40-0x00ffff
if ((o[0]==0)&&(o[1]==0xff)&&(o[2]&0xc0)) { // check if address 0x00ff40-0x00ffff
		fprintf(stderr,"WARNING! Bootloader protected at 0x00ff%02x (block %i)\n",o[2],ocnt);
		return;
}
// calcucate crc and output buffer to stdout
ocrc=0;
for(i=0;i<68;i++) {
	ocrc+=o[i];
	printf("%c",o[i]);
}
printf("%c",-ocrc);
ocnt++;
}

void outputhex(int addr, unsigned char c)
{
static int curaddr, baseaddr, outindex=0;

if (outstate==0) { // start a new block
	curaddr=addr;
	baseaddr=addr&~(0x3f);	// addr mod 64 = 0
	o[0]=(baseaddr>>16)&0xff;
	o[1]=(baseaddr>>8)&0xff;
	o[2]=(baseaddr)&0xff;
	o[3]=64;	
	// fill block with 0xff
	for(outindex=4;outindex<68;outindex++) o[outindex]=0xff;
	// skip any mod 64 bytes
	outindex=4+(curaddr&0x3f);
	outstate=1;
}
if (addr!=curaddr) { // new byte skipped addresses
	outindex=addr-baseaddr+4;
	if((outindex>=68)||(outindex<4)) { // not in same block
		outputhexblock();
		outstate=0;
		outputhex(addr,c);
		return;
	}
	curaddr=addr;	// same block, continue
}
// new byte at next address
o[outindex++]=c;
curaddr++;
if (outindex==68) {	// block full: output
	outputhexblock();
	outstate=0;
}
return;
}

int hex2bin(char b)
{
if ((b-=48)<0) return -1;
if (b<10) return b;
if ((b-=7)<10) return -1;
if (b>15) return -1;
return b;
}

int read2bhex(unsigned char *out) 
{
int a,b,m,l;
if ((a=getchar())==EOF) return 0;
if ((b=getchar())==EOF) return 0;
if ((m=hex2bin((char)a))==-1) return 0;
if ((l=hex2bin((char)b))==-1) return 0;
*out=(unsigned char) (m<<4) + (unsigned char) (l);
icrc+=*out;
return 1;
}

int read4bhex(unsigned short *out)
{
unsigned char m,l;
if (!(read2bhex(&m))) return 0;
if (!(read2bhex(&l))) return 0;
*out=(m<<8) + l;
return 1;
}


int main()
{
int line=0, instate=1, t, i;
unsigned int fulladdr=0;
unsigned short s, addr;
unsigned char c, len, type;

while(instate) {
	icrc=0;
	line++;
	if (getchar()!=':') {
		fprintf(stderr,"Input error on line %i (:)\n",line);
		return 1;
	}
	if (!(read2bhex(&len))) {
		fprintf(stderr,"Input error on line %i (len)\n",line);
		return 1;
	}
	if (!(read4bhex(&addr))) {
		fprintf(stderr,"Input error on line %i (addr)\n",line);
		return 1;
	}
	if (!(read2bhex(&type))) {
		fprintf(stderr,"Input error on line %i (type)\n",line);
		return 1;
	}
	//switch type
	switch(type) {
		case 0: // data
			fulladdr &= 0xffff0000;
			fulladdr |= addr;
			for(i=0;i<len;i++) {
				if (!(read2bhex(&c))) {
					fprintf(stderr,"Input error on line %i (data)\n",line);
					return 1;
				}
				outputhex(fulladdr,c);
				fulladdr++;
			}
			break;
		case 1: // end-of-file
			instate=0;
			break;
		case 2: // extended segment address
			fprintf(stderr,"Error! Unsupported hex type %i on line %i\n",type,line);
			return 1;
			break;
		case 4: // extended linear address
			if (!(read4bhex(&s))) {
				fprintf(stderr,"Input error on line %i (type)\n",line);
				return 1;
			}
			fulladdr = s<<16;
			break;
		default: // data record
			fprintf(stderr,"Error! Unknown hex type %i on line %i\n",type,line);
			return 1;
			break;
	}
	//crc check...
	if (!(read2bhex(&c))) {
		fprintf(stderr,"Input error on line %i\n",line);
		return 1;
	}
	if(icrc!=0) {
		fprintf(stderr,"CRC Error! Got %i on line %i\n",icrc,line);
		return 1;
	}
	// read until EOL
	if (instate) {
		t=0;
		while ((t!='\n') && (t!=EOF)) t=getchar();
	}

}
fprintf(stderr,"%i lines read\n",line);

// output remaining bytes (if any)
if (outstate!=0) outputhexblock();

// output start vector block (if needed)
if ((resetstate & BOOT_DONE_RESET)==0) {
	// set up block at 0x00ff00 using a dummy byte
	outputhex(0x0000ff38,0xff);
	outputhexblock();
}

// output reset vector block (if needed)
if ((resetstate & BOOT_HAVE_RESET)==0) {
	// set up block at 0x000000 using a dummy byte
	outputhex(0x00000000,0xff);
	outputhexblock();
}

fprintf(stderr,"OK! %i*69=%i bytes output\n",ocnt,ocnt*69);
return 0;
}
